/*	video.c
 *
 *	Video stream functions for motion.
 *	Copyright 2000 by Jeroen Vreeken (pe1rxq@amsat.org)
 *	This software is distributed under the GNU public license version 2
 *	See also the file 'COPYING'.
 *
 */
#include "motion.h"
#include "conf.h"
/* for the v4l stuff: */
#include "sys/mman.h"
#include "sys/ioctl.h"
#include "linux/videodev.h"
/* for the axis stuff: */
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define AXIS_WIDTH_1	320
#define AXIS_HEIGHT_1	240
#define AXIS_WIDTH_2	640
#define AXIS_HEIGHT_2	480
#define CONNECT_FAIL	1
#define DOWNLOAD_FAIL	2
#define DECOMPRESS_FAIL	4

extern struct config conf;


/*******************************************************************************************
	Video4linux capture routines
*/

int fmt=VIDEO_PALETTE_RGB24;
int read_img=0;

char *start_capture_v4l (int dev, int width, int height, int input, int norm)
{
        struct video_capability vid_caps;
	struct video_channel vid_chnl;
	struct video_window vid_win;
	struct video_mbuf vid_buf;
	char *map;

	if (ioctl (dev, VIDIOCGCAP, &vid_caps) == -1) {
		perror ("ioctl (VIDIOCGCAP)");
		return (NULL);
	}
	if (vid_caps.type & VID_TYPE_MONOCHROME) fmt=VIDEO_PALETTE_GREY;
	/* Set channels and norms, NOT TESTED my philips cam doesn't have a
	 * tuner. update: it works with my brothers haupage wintv (bttv driver)
	 */
        if (input != IN_DEFAULT) {
                vid_chnl.channel = -1;
                if (ioctl (dev, VIDIOCGCHAN, &vid_chnl) == -1) {
                        perror ("ioctl (VIDIOCGCHAN)");
                } else {
                        vid_chnl.channel = input;
                        vid_chnl.norm    = norm;
                        if (ioctl (dev, VIDIOCSCHAN, &vid_chnl) == -1) {
                                perror ("ioctl (VIDIOCSCHAN)");
                                return (NULL);
                        }
                }
        }
	if (ioctl (dev, VIDIOCGMBUF, &vid_buf) == -1) {
		fprintf(stderr, "no mmap falling back on read\n");
		if (ioctl (dev, VIDIOCGWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCGWIN");
			return (NULL);
		}
		vid_win.width=width;
		vid_win.height=height;
		if (ioctl (dev, VIDIOCSWIN, &vid_win)== -1) {
			perror ("ioctl VIDIOCSWIN");
			return (NULL);
		}
		read_img=1;
		map=malloc(width*height*3);
		return (map);
	}
	/* If we are going to capture greyscale we need room to blow the image up */
	if (fmt==VIDEO_PALETTE_GREY)
		map=mmap(0, vid_buf.size*3, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	else
		map=mmap(0, vid_buf.size, PROT_READ|PROT_WRITE, MAP_SHARED, dev, 0);
	
	if ((unsigned char *)-1 == (unsigned char *)map)
		return (NULL);
	return map;
}

char *next_capture_v4l (int dev, char *map, int width, int height)
{
	int i;
	char *grey, *rgb;
	struct video_mmap vid_mmap;

    	sigset_t    set, old;

	if (read_img) {
		if (fmt==VIDEO_PALETTE_GREY) {
			if (read(dev, map, width*height) != width*height)
				return NULL;
		} else {
			if (read(dev, map, width*height*3) != width*height*3)
				return NULL;
		}
	} else {
		vid_mmap.format=fmt;
		vid_mmap.frame=0;
		vid_mmap.width=width;
		vid_mmap.height=height;
    
        	sigemptyset (&set);	     //BTTV hates signals during IOCTL
	        sigaddset (&set, SIGCHLD);   //block SIGCHLD & SIGALRM
       		sigaddset (&set, SIGALRM);   //for the time of ioctls
        	sigprocmask (SIG_BLOCK, &set, &old);
                    
		if (ioctl(dev, VIDIOCMCAPTURE, &vid_mmap) == -1) {
	        	sigprocmask (SIG_UNBLOCK, &old, NULL);
			return (NULL);
		}
		if (ioctl(dev, VIDIOCSYNC, &vid_mmap) == -1) {
	        	sigprocmask (SIG_UNBLOCK, &old, NULL);
			return (NULL);
		}
		
        	sigprocmask (SIG_UNBLOCK, &old, NULL); //undo the signal blocking
	}
	/* Blow up a grey */
	if (fmt==VIDEO_PALETTE_GREY) {
		i=width*height;
		grey=map+i-1;
		rgb=map+i*3;
		for (; i>=0; i--, grey--) {
			*(rgb--)=*grey;
			*(rgb--)=*grey;
			*(rgb--)=*grey;
		}
	}
	return map;
}

int start_videopipe_v4l (int dev, int width, int height)
{
	struct video_picture vid_pic;
	struct video_window vid_win;
	
	if (ioctl(dev, VIDIOCGPICT, &vid_pic)== -1) {
		perror("ioctl VIDIOCGPICT");
		return(1);
	}
	vid_pic.palette=VIDEO_PALETTE_RGB24;
	if (ioctl(dev, VIDIOCSPICT, &vid_pic)== -1) {
		perror("ioctl VIDIOCSPICT");
		return(1);
	}
	if (ioctl(dev, VIDIOCGWIN, &vid_win)== -1) {
		perror("ioctl VIDIOCGWIN");
		return(1);
	}
	vid_win.height=height;
	vid_win.width=width;
	if (ioctl(dev, VIDIOCSWIN, &vid_win)== -1) {
		perror("ioctl VIDIOCSWIN");
		return(1);
	}
	return 0;
}

int put_videopipe_v4l (int dev, char *image, int height, int width)
{
	return write(dev, image, height*width*3);
}

/*******************************************************************************************
	Axis 2100 capture routines
*/

/*  This is a driver for Motion version 1.6 for the Axis 2100 and similar cameras.
    Copyright (C) 2000  Tristan Willy

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


int connect_http(void);
FILE *download_jpg(void);
unsigned char *decompress_jpeg();

/* Global Vars. */
struct sockaddr_in address;
int socketfd;
char *get_command;


/* API */


/* Things needed to be included in this function. 
 *    - Setup the IP structers
 *    - Check to make sure the width and height are correct
 *    - Ping to make sure the device is reachable. (NOT DONE YET)
 */ 
char *start_capture_axis (int dev, int width, int height, int input, int norm){
	char *pict_ptr;

	/* Just making sure... */
	if((width != AXIS_WIDTH_1 || height != AXIS_HEIGHT_1) && (width != AXIS_WIDTH_2 || height != AXIS_HEIGHT_2)){
		printf("Error: the width and height are incorrect. An abort is neccessary. ABORT...\n");
		return NULL;
	}

	/* Determine what CGI-Script is used for the resolution. */
	if((get_command = malloc(45)) == NULL){  // Just a bit more than needed.
		perror("driver.c");
		return NULL;
	}
	
	if(width == AXIS_WIDTH_1)
		strcpy(get_command, "GET /cgi-bin/fullsize.jpg HTTP/1.0\r\n\r\n");
	else
		strcpy(get_command, "GET /cgi-bin/hugesize.jpg HTTP/1.0\r\n\r\n");

		
	/* Make enough memory, 3 bytes per pixel (RGB - 8bit) */
	if((pict_ptr = malloc(width * height * 3)) == NULL){
		perror("driver.c");
		return NULL;
	}

	/* Set up the address */
	address.sin_family = AF_INET;
	address.sin_addr.s_addr = inet_addr(conf.axis);
	address.sin_port = htons(80);


	return pict_ptr;
}

/* Captures the next frame. MUST RETURN A FRAME OTHERWISE THE
 * PROGRAM EXITS! */
char *next_capture_axis (int dev, char *map, int width, int height){
	FILE *jpeg;
	unsigned char *uncompressed;
	int flag = 1;  // Start initial loop

	while(flag){
		flag = 0;

		/* Make the socket. */
		socketfd = socket(AF_INET, SOCK_STREAM, 0);

		if(connect_http()){
			close(socketfd);
			flag++;
		}

		if((jpeg = download_jpg()) == NULL && !flag){
			close(socketfd);
			flag++;
		}  	
		
		/* Decompress the jpeg. */
		if((uncompressed = decompress_jpeg(jpeg)) == NULL && !flag){
			fclose(jpeg);
			flag++;
		}
	}

	/* Now copy the image over. */
	memcpy(map, uncompressed, width * height * 3);

	/* Cleanup */
	free(uncompressed);
	fclose(jpeg);	

	return map;     /* a good memory location with a valid address */
}


/* COMPONENTS */

/* Connect and then release. */
int connect_http(){

	if((connect(socketfd, (struct sockaddr *)&address, sizeof(address))) == -1){
		perror("Connect");
		return 1;
	}

	write(socketfd, get_command, strlen(get_command));

	return 0;
}

/* Get information from the server until 3 newlines are passed.
 * After that then save the data to a file. 
 */
FILE *download_jpg(void){
	ssize_t ret, i = 0;
	char buff[512];
	FILE *jpeg;
	int newlines = 0;


	if((jpeg = tmpfile()) == NULL)
		return NULL;

	while((ret = read(socketfd, buff, 512))){
		if(newlines < 3)
			for(i = 0; i < ret && newlines < 3; i++)
				 if(buff[i] == '\n')
				 	newlines++;
				 	
		write(fileno(jpeg), buff + i, ret - i);
		i = 0;
	}

	rewind(jpeg);
	close(socketfd);
	return jpeg;
}


/* Decompress the jpeg and turn it into a BGR (1*1*1) pointer. */
unsigned char *decompress_jpeg(FILE *jpeg){
	struct jpeg_decompress_struct cinfo;
	struct jpeg_error_mgr jerr;
	unsigned char *line, *pic, *old_pic;
	JSAMPROW row[1];
	unsigned int i, line_size;

	cinfo.err = jpeg_std_error(&jerr);
	jpeg_create_decompress(&cinfo);

	jpeg_stdio_src(&cinfo, jpeg);

	jpeg_read_header(&cinfo, TRUE);

	cinfo.out_color_space = JCS_RGB;
	
	jpeg_start_decompress(&cinfo);
	line_size = cinfo.output_width * 3;


	if((line = malloc(cinfo.output_width * 3)) == NULL)
		return NULL;

	if((pic = malloc(cinfo.output_width * cinfo.output_height * 3)) == NULL){
		free(line);
		return NULL;
	}

	old_pic = pic;
	row[0] = line;
	while(cinfo.output_scanline < cinfo.output_height){
		jpeg_read_scanlines(&cinfo, row, 1);

		for(i = 0; i < line_size; i += 3){  // Inverse the image bytes.
			pic[i] = line[i+2];
			pic[i+1] = line[i+1];
			pic[i+2] = line[i];
		}
		pic += line_size;
	}

	jpeg_finish_decompress(&cinfo);
	jpeg_destroy_decompress(&cinfo);
	
	return old_pic;
}



/*******************************************************************************************
	Wrappers calling the actual capture routines
*/
char *start_capture (int dev, int width, int height, int input, int norm)
{
	if (!conf.axis) return start_capture_v4l (dev, width, height, input, norm);
	else return start_capture_axis (dev, width, height, input, norm);
}

char *next_capture (int dev, char *map, int width, int height)
{
	if (!conf.axis) return next_capture_v4l (dev, map, width, height);
	else return next_capture_axis (dev, map, width, height);
}

int start_videopipe (int dev, int width, int height)
{
	return start_videopipe_v4l (dev, width, height);
}

int put_videopipe (int dev, char *image, int height, int width)
{
	return put_videopipe_v4l (dev, image, height, width);
}
